Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path.
Require Import pred prelude idynamic ordtype finmap pcm coding unionmap heap.
Require Import feap state zigzag world inject rely hide atomic autosplit.
Require Import privstate footprint model lemmas stsep extra_lemmas histories. 
Require Import getter treiber alloc.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(* A demonstration of the frame rule *)

Section TreiberPopFramed.
Variable T : treiberT.
Notation A := (tb_set T). 
Notation sent := (tb_sent T).
Notation tb := (tb_label T).
Notation st := [encoded_set of seq A].
Notation coh := (tb_coh T).
Notation U := (tbU T).
Notation tb_getso s := (getso (tbg T) s).
Notation tb_self s := (self (tbg T) s).
Notation tb_other s := (other (tbg T) s).
Notation tb_joint s := (joint (tbg T) s).
Notation guar := (Mod.internal (Treiber T)). 
Notation W := (Treiber T).

Definition pop_framed_tp :=
  {h hs}, STsep [W] (fun i => tb_self i = hs /\ [h <<= hs \+ tb_other i],
                     fun v m =>  
                      [/\ [h <<= hs \+ tb_other m] &
                          if v is Some e then 
                            exists t ls, [/\ last_key h < t & 
                                             tb_self m = hs \+ t \\-> (e :: ls, ls)]
                          else 
                            exists t, [/\ last_key h <= t, 
                                          tb_self m = hs & 
                                          [[::] in hs \+ tb_other m at t]]]).

Program Definition pop_framed : pop_framed_tp := Do (pop T).
Next Obligation.
apply: gh=>i [h _] /= [<- H] C.
move: (cohVSO C) => /= Vso; move: (histK H) => /= K.
have X : sself [:: tb_getcoh T] i = Unit \+ tb_self i by rewrite unitL.
apply: frame_do X _ => //; apply: (gh_ex (tb_self i \+ tb_other i)).
apply: val_do=>[|x m [H1 H2] Cm o /= M]; rewrite !getE //. 
move: M H1 H2=>-> H1 H2; split; first by apply: hist_trans H H1.
case: x H2=>[a [t][ls][H2 ->]|[t][H2 -> H3]]; exists t; last first.
- by rewrite unitL; split=>//; apply: leq_trans H2.
by exists ls; rewrite (joinC (t \\-> _)); split=>//; apply: leq_ltn_trans H2. 
Qed.

End TreiberPopFramed.

Section TreiberPushFramed.
Variables (pv al : nat) (T : treiberT) (asent alk : ptr).
Notation A := (tb_set T). 
Notation sent := (tb_sent T).
Notation tb := (tb_label T).
Notation st := [encoded_set of seq A].
Notation W := (entangle (Priv pv) (Treiber T)). 
Notation guar := (Mod.internal W). 
Notation U := (tbU T). 
Notation tb_getso s := (getso (tbg T) s).
Notation tb_self s := (self (tbg T) s).
Notation tb_other s := (other (tbg T) s).
Notation tb_joint s := (joint (tbg T) s). 
Notation pv_getso s := (getso (pvg pv) s).
Notation pv_self s := (self (pvg pv) s).
Notation pv_other s := (other (pvg pv) s).
Notation pv_coh := (priv_coh pv).
Notation tb_coh := (TreiberCoh.Exports.tb_coh T).
Notation pt_coh := (Mod.coh W).

(* Use a concrete, CAS-based allocator *)
Notation AllocM := (AllocCAS al asent alk). 
Notation V := (entangle W AllocM). 
Notation al_coh := (al_coh al asent alk). 

Definition push_framed_tp e :=
  {h hs hS}, STsep [V] (fun i => [/\ tb_self i = hs, pv_self i = hS & [h <<= hs \+ tb_other i]],
                     fun (_ : unit) m => exists t ls, 
                       [/\ tb_self m = hs \+ t \\-> (ls, e :: ls), pv_self m = hS,
                           [h <<= hs \+ tb_other m] &
                           last_key h < t]). 

Program Definition push_framed e : push_framed_tp e := Do (push pv al asent alk e).
Next Obligation.
apply: gh=>i [h [_ hS]] /= [<-]Hp H Ci; rewrite starAC starC in Ci.
case: Ci (Ci) H Hp=>_ [xt][s][->{i}] Ct Cpa Ci H Hp.
move: (cohVSO Ct) => /= Vso. 
case: Cpa (Cpa) Ci H Hp=>_[sp][sa][->{s}]Cp Ca Cpa Ci H Hp.
rewrite !(getC Ct Ci) !(getC Cp Ci) /= in H Hp *. 
have X : tb_self xt = Unit \+ tb_self xt by rewrite unitL.
have Y: Mod.coh V = star_coh tb_coh (star_coh pv_coh al_coh).
- by rewrite /= starAC starC.
apply: frame' Ct X Y _.
apply: (gh_ex (tb_self xt \+ tb_other xt)); apply:(gh_ex (pv_self sp)).
apply: val_do=>[/cohV V|_].
- by rewrite !getE // (getCV Cp).
move=>m [t][ls][H1 Hp2 H2] K Cm xt2 y o Em.
move: Em H1 Hp2 H2 Cm =>->{m} H1 Hp2 H2 Ci2 Ct2 Cy Ht.
rewrite !(getC Ct2 Ci2) /= in H1 H2; rewrite {}Ht in H2 *.
have V: valid (tb ->> [tb_self xt2 \+ tb_self xt, tb_joint xt2, o] \+ y).
- by move/cohV: Ci2; rewrite {1}(cohE Ct2) /valid /= !fsE !fvalidPtUn. 
exists t, ls; rewrite !getE // {1}joinC -H1; split=>//.
- case: Cy Ci2 V Hp2=>_ [s1][s2][->] C1 C2 /cohV V' V.
  by rewrite !(getCV C1) //= Hp. 
- by apply: hist_trans H H2.
by apply: leq_ltn_trans K; apply: histK H.
Qed.

End TreiberPushFramed.

